
local AbsorbsMonitor = _G.AbsorbsMonitor;

AbsorbsViewer = LibStub("AceAddon-3.0"):NewAddon("AbsorbsViewer", "AceConsole-3.0", "AceEvent-3.0");

local SM = LibStub:GetLibrary("LibSharedMedia-3.0");

local AbsorbEffects;
local AbsorbEffectNames;

local baseFrame;
local framePool_Units = {};
local framePool_Effects = {};
local activeUnitsMap = {};
local activeUnitsList = {};
local lowestUnitFrame = nil;

local isHidden;


-- #################################################################################################


local function GetClassColorRGB(class)
	local r, g, b;

	if _G.RAID_CLASS_COLORS and _G.RAID_CLASS_COLORS[class] then
		r, g, b = _G.RAID_CLASS_COLORS[class].r, _G.RAID_CLASS_COLORS[class].g, _G.RAID_CLASS_COLORS[class].b
	else
		r, g, b = 0.63, 0.63, 0.63;	
	end
	
	return r, g, b;
end

local function OnDragStart(self)
	if IsAltKeyDown() then
		self.isMoving = true;
		self:StartMoving();
	end
end

local function OnDragStop(self)
	if self.isMoving then
		self.isMoving = false;
		self:StopMovingOrSizing();

		local x, y = self:GetLeft(), self:GetTop();
	end
end

local function OnEnter(self)
	GameTooltip:SetOwner(self, "ANCHOR_TOPLEFT");
	GameTooltip:SetText("ALT + Drag to move the frame anchor", nil, nil, nil, nil, 1);
end

local function OnLeave(self)
	GameTooltip:Hide();	
end

local backdrop = {bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
		edgeFile = "Interface\\ChatFrame\\ChatFrameBackground", edgeSize = 0.90,
		insets = {left = 1, right = 1, top = 1, bottom = 1}};
		
-- Note this function is exactly INVERSE to SortEffects in Core.lua
local function SortEffects(a, b)
	if a.priority == b.priority then
		return (a.amount > b.amount);
	else
		return (a.priority < b.priority);
	end
end

local function RepositionUnits()
	for i, frame in pairs(activeUnitsList) do
		frame:ClearAllPoints();
		
		if(i > 1) then
			frame:SetPoint("TOPLEFT", activeUnitsList[i - 1], "BOTTOMLEFT", 0, -2);
		else
			frame:SetPoint("TOPLEFT", baseFrame, "BOTTOMLEFT", 0, -2);
		end
	end
	
	lowestUnitFrame = activeUnitsList[#activeUnitsList];
end

local function RepositionEffects(guid)
	if(not activeUnitsMap[guid]) then return; end
	
	local effectsList = activeUnitsMap[guid].effectsList;
	
	table.sort(effectsList, SortEffects);
	
	for i, frame in pairs(effectsList) do
		frame:ClearAllPoints();
	
		if(i > 1) then
			frame:SetPoint("TOPLEFT", effectsList[i - 1], "BOTTOMLEFT", 0, -2);
		else
			frame:SetPoint("TOPLEFT", activeUnitsMap[guid], "TOPLEFT", 10, -22 );
		end
	end
end

local function CreateBaseFrame()
	baseFrame = CreateFrame("Frame", "absorbsview_baseFrame", UIParent);
	
	baseFrame:SetHeight(20);
	baseFrame:SetMovable(true);
	baseFrame:EnableMouse(true);
	baseFrame:SetClampedToScreen(true);
	baseFrame:RegisterForDrag("LeftButton");
	baseFrame:SetBackdrop(backdrop);
	baseFrame:SetBackdropColor(0, 0, 0, 0.5);
	baseFrame:SetBackdropBorderColor(1, 1, 1, 0.6);
	baseFrame:SetScript("OnDragStart", OnDragStart);
	baseFrame:SetScript("OnDragStop", OnDragStop);
	baseFrame:SetScript("OnShow", OnShow);
	baseFrame:SetScript("OnEnter", OnEnter);
	baseFrame:SetScript("OnLeave", OnLeave);
	baseFrame.name = "absorbs_viewer";
	
	baseFrame.text = baseFrame:CreateFontString(nil, "OVERLAY");
	baseFrame.text:SetPoint("LEFT", baseFrame, 10, 0);
	baseFrame.text:SetFontObject(GameFontHighlightSmall);
	baseFrame.text:SetText("AbsorbsMonitor Effect Viewer");
	
	baseFrame:SetScale(1.0);
	
	baseFrame:SetWidth(225);	
	
	baseFrame:SetAlpha(1.0);
	
	baseFrame:ClearAllPoints();
	baseFrame:SetPoint("CENTER", UIParent, "CENTER");
	
	baseFrame.CloseButton = CreateFrame("Button", nil, baseFrame);
	baseFrame.CloseButton:SetNormalTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Up.blp");
	baseFrame.CloseButton:SetPushedTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Down.blp");
	baseFrame.CloseButton:SetHighlightTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight.blp");
	baseFrame.CloseButton:SetWidth(20);
	baseFrame.CloseButton:SetHeight(20);
	baseFrame.CloseButton:SetPoint("TOPRIGHT", baseFrame, "TOPRIGHT", 0, 0);
	baseFrame.CloseButton:SetScript("OnClick", function(self) self:GetParent():Hide(); isHidden = true; end);
	
	lowestUnitFrame = baseFrame;
end

local function GetUnitFrame()
	if(#framePool_Units > 0) then
		return table.remove(framePool_Units, 1);
	end
	
	local frame = CreateFrame("Frame", nil, baseFrame);

	--frame:SetMovable(true);
	--frame:EnableMouse(true);
	--frame:SetClampedToScreen(true);
	--frame:RegisterForDrag("LeftButton");
	frame:SetBackdrop(backdrop);
	frame:SetBackdropColor(0, 0, 0, 0.5);
	frame:SetBackdropBorderColor(1, 1, 1, 0.3);
	--frame:SetScript("OnMouseDown", StaticBarGroup_OnMouseDown);
	--frame:SetScript("OnDragStart", StaticBarGroup_OnDragStart);
	--frame:SetScript("OnDragStop", StaticBarGroup_OnDragStop);
	--frame:SetScript("OnShow", OnShow);
	--frame:SetScript("OnEnter", StaticBarGroup_OnEnter);
	--frame:SetScript("OnLeave", StaticBarGroup_OnLeave);
	--frame.name = name;
	
	local path, size, style = GameFontHighlight:GetFont();	
	--size = size;	
	
	frame.text = frame:CreateFontString(nil, "OVERLAY");
	frame.text:SetPoint("TOPLEFT", frame, 10, -5);
	frame.text:SetFontObject(GameFontHighlightSmall);
	frame.text:SetFont(path, size, style);
	
	frame:SetScale(1.0);
	
	frame:SetWidth(225);	
	
	--frame:EnableMouse(true);
	frame:SetAlpha(1.0);
	
	frame:ClearAllPoints();
	
	return frame;
end

local function RemoveUnit(guid)
	if(not activeUnitsMap[guid]) then return; end
	
	local unitFrame = activeUnitsMap[guid];
	
	activeUnitsMap[guid] = nil;
	
	for i, frame in pairs(activeUnitsList) do
		if(frame == unitFrame) then
			table.remove(activeUnitsList, i);
			break;
		end
	end
		
	unitFrame.effectsMap = nil;
	unitFrame.effectsList = nil;
	
	unitFrame:Hide();
	table.insert(framePool_Units, unitFrame);
	
	
	if(#activeUnitsList > 0) then
		RepositionUnits();
	else
		lowestUnitFrame = baseFrame;
	end
end

local function GetEffectFrame()
	if(#framePool_Effects > 0) then
		return table.remove(framePool_Effects, 1);
	end
		
	local frame = CreateFrame("StatusBar", nil, baseFrame);
	frame:Hide();		
	
	frame:SetClampedToScreen(true);
	frame:SetMinMaxValues(0, 1);
	
	local texture = SM:Fetch(SM.MediaType.STATUSBAR, "BantoBar");	
		
	frame.bg = CreateFrame("StatusBar", nil, frame);
	frame.bg:SetMinMaxValues(0, 1);
	frame.bg:SetValue(1.0);
	frame.bg:SetPoint("TOPLEFT", frame, "TOPLEFT", 0, 0);
	frame.bg:SetFrameLevel(0);
	frame.bg:SetStatusBarTexture(texture);
	frame.bg:SetWidth(205);
	frame.bg:SetHeight(14);
	frame.bg:SetStatusBarColor(0.5, 0.5, 0.5, 0.5);
	
	frame:SetStatusBarTexture(texture);	
	frame:SetWidth(205);
	frame:SetHeight(14);
	frame:SetScale(1.0);	
	
	local path, size, style = GameFontHighlight:GetFont();	
	size = size - 2;
	
	frame.textName = frame:CreateFontString(nil, "OVERLAY");
	frame.textName:SetFontObject(GameFontHighlight);
	frame.textName:SetJustifyH("LEFT");
	frame.textName:SetPoint("LEFT", frame, "LEFT", 0, 0);
	frame.textName:SetFont(path, size, style);	
	frame.textName:SetHeight(14);
	frame.textName:SetWidth(125);	
			
	frame.textAmount = frame:CreateFontString(nil, "OVERLAY");
	frame.textAmount:SetFontObject(GameFontHighlight);
	frame.textAmount:SetJustifyH("RIGHT");
	frame.textAmount:SetPoint("RIGHT", frame, "RIGHT", -3, 0);
	frame.textAmount:SetFont(path, size, style);
	frame.textAmount:SetHeight(10);
	frame.textAmount:SetWidth(80);
		
	frame:SetAlpha(1.0);
	
	return frame;
end

local function ReleaseEffectFrame(frame)
	frame:Hide();
	
	table.insert(framePool_Effects, frame);
end

local function AddEffect(guid, unitName, spellId, amount, quality)
	local unitFrame, effectFrame;
	local r, g, b;

	if(not activeUnitsMap[guid]) then
		unitFrame = GetUnitFrame();
		
		unitFrame:SetHeight(30);
		
		local _, unitClass = UnitClass(unitName);
		r, g, b = GetClassColorRGB(unitClass);
		
		unitFrame.text:SetText(unitName);
		unitFrame.text:SetTextColor(r, g, b);
		
		activeUnitsMap[guid] = unitFrame;
		table.insert(activeUnitsList, unitFrame);
		
		unitFrame:SetPoint("TOPLEFT", lowestUnitFrame, "BOTTOMLEFT", 0, -2);
		lowestUnitFrame = unitFrame;
		
		unitFrame:Show();
	else
		unitFrame = activeUnitsMap[guid];
	end
	
	effectFrame = GetEffectFrame();
	
	effectFrame.spellId = spellId;
	
	effectFrame:SetValue(1.0);
	effectFrame.amount = amount;
	effectFrame.max = amount;
	effectFrame.priority = AbsorbEffects[spellId][1];
	
	r = 1 - quality;
	g = quality;
	
	effectFrame:SetStatusBarColor(r, g, 0, 1.0);
	
	if(AbsorbEffectNames[spellId]) then
		effectFrame.textName:SetText(AbsorbEffectNames[spellId]);
	else
		effectFrame.textName:SetText("SpellId "..spellId);
	end
	
	effectFrame.textAmount:SetText(amount);
	
	if(not unitFrame.effectsMap) then
		effectFrame:SetPoint("TOPLEFT", unitFrame, "TOPLEFT", 10, -22);
				
		unitFrame.effectsMap = {[spellId] = effectFrame};
		unitFrame.effectsList = {effectFrame};
	else		
		unitFrame.effectsMap[spellId] = effectFrame;			
		table.insert(unitFrame.effectsList, effectFrame);
		
		RepositionEffects(guid);
	end		
		
	unitFrame:SetHeight(unitFrame:GetHeight() + 14);
	effectFrame:Show();
end

local function RemoveEffect(guid, spellId)
	if(activeUnitsMap[guid] and activeUnitsMap[guid].effectsMap[spellId]) then
		local unitFrame = activeUnitsMap[guid];
	
		ReleaseEffectFrame(unitFrame.effectsMap[spellId]);
		
		unitFrame.effectsMap[spellId] = nil;
		
		for i, frame in pairs(unitFrame.effectsList) do
			if(frame.spellId == spellId) then
				table.remove(unitFrame.effectsList, i);
				break;
			end
		end
		
		if(#(unitFrame.effectsList) == 0) then			
			RemoveUnit(guid);
		else
			RepositionEffects(guid);
			unitFrame:SetHeight(unitFrame:GetHeight() - 14);
		end
	end		
end

local function UpdateEffect(guid, spellId, amount)
	if(amount == 0) then
		RemoveEffect(guid, spellId);
	end

	if(activeUnitsMap[guid] and activeUnitsMap[guid].effectsMap[spellId]) then
		local effectFrame = activeUnitsMap[guid].effectsMap[spellId];
		
		if(amount >= effectFrame.max) then
			effectFrame:SetValue(1.0);
			effectFrame.max = amount;
			
			-- We might need to reposition now
			RepositionEffects(guid);
		else
			effectFrame:SetValue(amount / effectFrame.max);
		end
		
		effectFrame.textAmount:SetText(amount);
	end
end


-- #################################################################################################


function AbsorbsViewer:Absorbs_Unit_NewEffect(obj, sourceGUID, sourceName, destGUID, destName, spellId, amount, quality)
	AddEffect(destGUID, destName, spellId, amount, quality);
end

function AbsorbsViewer:Absorbs_Unit_UpdatedEffect(obj, guid, spellId, amount)
	UpdateEffect(guid, spellId, amount);
end

function AbsorbsViewer:Absorbs_Unit_RemovedEffect(obj, guid, spellId)
	RemoveEffect(guid, spellId);
end

function AbsorbsViewer:Absorbs_Unit_Cleared(obj, guid)
	if(activeUnitsMap[guid]) then
		-- no need to worry about any data structures here
		for i, frame in pairs(activeUnitsMap[guid].effectsList) do
			ReleaseEffectsFrame(frame);
		end
		
		RemoveUnit(guid);
	end
end

function AbsorbsViewer:OnSlashCommand()
	baseFrame:Show();
	isHidden = false;
end

function AbsorbsViewer:OnEnable()
	CreateBaseFrame();
	baseFrame:Hide();
	isHidden = true;
	
	AbsorbEffects = AbsorbsMonitor.Effects;
	AbsorbEffectNames = AbsorbsMonitor.EffectNames;
	
	self:RegisterMessage("Absorbs_Unit_NewEffect");
	self:RegisterMessage("Absorbs_Unit_UpdatedEffect");
	self:RegisterMessage("Absorbs_Unit_RemovedEffect");
	self:RegisterMessage("Absorbs_Unit_Cleared");
	
	self:RegisterChatCommand("av", "OnSlashCommand");
end

function AbsorbsViewer:OnDisable()
end
